#!/usr/bin/ruby
# Proxy.rby -- the proxy module for frox test suite. This file contains
# support code for running the proxy, and communicating with the client
# module to request files. Also contains the definition for the base
# class of the test suite so that test.rby can contain tests only.


# Frox objects represent a particular configuration of frox which may be
# run many times. The configuration includes the frox binary (which may
# have been compiled with ./configure options), and a config file which
# is normally generated by this class based on /etc/froxtester.conf and
# any other requests.
class Frox
  attr_reader :cprts, :pprts, :aprts
  def initialize(exec, config = nil)
    @exec = exec
    @config = config
    @cprts = [40000, 40999]
    @pprts = [41000, 41999]
    @aprts = [42000, 42999]
  end
  def write_config
    File.open(@config, "w") { |file|
      file.puts "Listen " + FroxDefaults::Listen
      file.puts "Port " + FroxDefaults::Port.to_s
      file.puts "WorkingDir " + FroxDefaults::WorkingDir
      file.puts "User " + FroxDefaults::User
      file.puts "Group " + FroxDefaults::Group
      file.puts "ResolvLoadHack " + FroxDefaults::ResolvLoadHack
      file.puts "LogFile " + FroxDefaults::WorkingDir + "/frox.log"
      #file.puts "LogFile stderr"
      #file.puts "LogLevel 100"
      file.puts "XferLogging yes"
      file.puts "ControlPorts #{@cprts[0]}-#{@cprts[1]}"
      file.puts "PassivePorts #{@pprts[0]}-#{@pprts[1]}"
      file.puts "ActivePorts #{@aprts[0]}-#{@aprts[1]}"
      file.puts "DoNTP yes"
      file.puts "MaxForks 10"

      file.puts @configadd if @configadd
      file.puts "ACL Allow * - *"
    }
  end
  def add_config(str)
    @configadd = str if !@configadd
    @configadd += "\n" + str if @configadd
  end
  def run
    write_config
    @logfilefd = File.open(FroxDefaults::WorkingDir + '/frox.log')
    purgeLog
    exec(@exec + (@config ? (" -f " + @config) : "")) if fork==nil
    a = waitForLog(/Listening on/, /ERROR/)
    run if a.index('ERROR') # bind:address is already in use
  end
  def purgeLog
    @logfilefd.each_line do |line| end
  end
  def waitForLog(*regexp)
    loop do
      a = findLog(*regexp)
      return a if a
      sleep 0.05
    end
  end
  def findLog(*regexp)
    @logfilefd.each_line do |line|
      regexp.each do |r|
        return line if line=~r
      end
    end
    return false
  end
  def kill
    system("killall #{@exec}")
    Process.wait
    @configadd = ""
    @logfilefd.close if @logfilefd
  end
  def restart(purgecache = false)
    system("killall #{@exec}")
    Process.wait
    @logfilefd.close if @logfilefd
    system('rm `find ' + FroxDefaults::WorkingDir + '/cache/ -type f` 2>/dev/null') \
      if purgecache
    run
  end
end

class FroxException < RuntimeError
end

class ClientDriver
  attr_reader :froxlogin
  attr_reader :pasvaddr
  attr_reader :portconnect
  attr_reader :md5sum

  def initialize(ip, exec)
    @ip = ip
    @exec = exec
  end
  def run
    #cmd = "ssh " + @ip + " " + @exec
    @client = IO.popen(@exec, "w+")
  end
  def connect(ip, port, user="ftp", pass="ftp", proxyip=nil, proxyport=2121)
    @client.puts "CONNECT " + ip + " " + port.to_s if !proxyip
    @client.puts "CONNECT " + proxyip + " " + proxyport.to_s if proxyip
    reply=""
    begin
      reply = @client.gets.split
      if reply[0] == "froxlogin"
        @froxlogin = (reply[1]=="true")
      end
    end until (reply[0]=="OK" || reply[0]=="FAIL")
    raise FroxException, "Unable to connect" if reply[0]=="FAIL"

    @client.puts "LOGIN " + user + " " + pass if !proxyip
    @client.puts "LOGIN "+user+"@"+ip+":"+port.to_s+" "+pass if proxyip
    reply = @client.gets.split
    if reply[0]=="FAIL"
      logout
      raise FroxException, "Unable to login"
    end
  end
  def passive=(setting)
    @client.puts "PASSIVE " + (setting ? "ON" : "OFF")
    @client.gets
  end
  def cd(dir)
    @client.puts "CD " + dir
    reply = @client.gets.split
    raise FroxException, "Unable to chdir" if reply[0]=="FAIL"
  end
  def retrieve(filename, rests=nil)
    @portconnect=nil
    @pasvaddr=nil
    @client.puts "RETR " + filename
    reply=""
    begin
      reply = @client.gets.split
      case reply[0]
        when "pasvaddr" then @pasvaddr = reply[1].split(':')
	when "portconnect" then @portconnect = reply[1].split(':')
	when "md5sum" then @md5sum = reply[1]
	when "FAIL" then raise FroxException, "Failed to retrieve file " + filename
      end
    end until (reply[0]=="OK")
  end
  def keep
    @client.puts "KEEP"
    ret = @client.gets.split[1]
    @client.gets
    return ret
  end
  def logout
    @client.puts "BYE"
    @client.gets
  end
  def quit
    @client.puts "QUIT"
  end
end

class FroxTester

  def initialize(proxy, clientip=TesterConfig::ClientIP,
                 clientexec=TesterConfig::ClientExec)
    @proxy = proxy
    @client = ClientDriver.new(clientip, clientexec)
    @client.run
    @passive = false
  end
  def passive(value)
    @client.passive = value
    @passive = value
  end
  def connect
    @total_connects +=1
    serverPort = @currentServer['Port']
    serverUser = @currentServer['User']
    serverPass = @currentServer['Pass']
    if @currentServer['NTP']
      proxyIP = TesterConfig::FroxIP
      proxyPort = 2121
    else
      proxyIP = nil
    end
    serverPort = 21 if !serverPort
    serverUser = "ftp" if !serverUser
    serverPass = "froxtester@localhost" if !serverPass
    @client.connect(@currentServer['Host'], serverPort, serverUser, serverPass,
                    proxyIP, proxyPort)
  end
  def logout
    @client.logout
  end
  def cd(dir)
    @client.cd(dir)
  end
  def retrieve(filename, md5sum=nil, expectedtofail = false)
    @total_retrieves +=1
    @client.retrieve(filename)
    raise FroxException, "md5sum mismatch #{@client.keep}" \
                                         if md5sum && @client.md5sum!=md5sum
    if @passive && (@client.pasvaddr[1].to_i<@proxy.pprts[0] ||
                    @client.pasvaddr[1].to_i>@proxy.pprts[1])
      raise FroxException, "passive port out of range"
    elsif !@passive && (@client.portconnect[1].to_i<@proxy.aprts[0] ||
                        @client.portconnect[1].to_i>@proxy.aprts[1])
      raise FroxException, "active port out of range"
    end
  rescue
    raise if !expectedtofail
  else
    raise FroxException, "Retrieved nonexistant file" if expectedtofail
  end

  def quit
    @client.quit
  end

# Run the entire test suite defined by the modules that have been extended
# into this class.
  def run
    @total_tests = 0
    @total_connects = 0
    @total_retrieves = 0
    @errors = Array.new
    special_configs = Array.new
    configs = Array.new
    @tests = []
    methods.each {|method|
      @tests << method if method =~ /^test_/
      if method =~ /^config_/
        if method.split('_')[1]=='x'
	  special_configs << method
	else
          conf_level = method.split('_')[1].to_i
	  configs[conf_level] = [nil] if !configs[conf_level]
	  configs[conf_level] << method
	end
      end
    }
    config_levels = configs.size
    puts "  Servers to test against:"
    $FtpServers.each_index {|i|
      puts "       #{i+1}. #{$FtpServers[i]['Name']}"
    }
    config_level_counts = Array.new(config_levels, 0)
    firstLoop = true
    loop {
      if !firstLoop
        @proxy.kill
      end
      firstLoop = false

     # $stdout.write "#{config_level_counts.inspect}:  "
      configs_to_use = Array.new
      description = ''
      configs.each_index do |i|
        c = configs[i][config_level_counts[i]]
        configs_to_use << c if c
        description += "#{c.sub(/config_([0-9])+_/, '').sub(/ONLY_.*_ONLY_/, '')} " if c
      end
      description = description.split(' ').reverse.join(', ')
      description = 'Standard Config' if description.empty?

      run_tests(description, configs_to_use)

      #Increment counters:
      config_level_counts.each_index do |i|
        config_level_counts[i] += 1
	if config_level_counts[i] == configs[i].size
	  config_level_counts[i] = 0 if i!=config_levels-1
	else
	  break
	end
      end
      break if config_level_counts.last == configs.last.size
    }

    special_configs.each do |c|
      @proxy.kill
      run_tests(c.sub('config_x_', '').sub(/ONLY_.*_ONLY_/, ''), [ c ])
    end
    
    return @errors, @total_tests, @total_connects, @total_retrieves
  end

  def run_tests(description, configs_to_use)
    puts "  Using #{description}:"
    configs_to_use.each do |c|
      send(c, @proxy) if c
    end
    @proxy.run
    @tests.each {|test|
      testOnly = test.split('_ONLY_')[1]
      testDescrip = test.sub('test_', '').sub(/ONLY_.*_ONLY_/, '')
      next unless !testOnly || description.index(testOnly) 
      configProhibits = false
      configs_to_use.each do |c|
        configOnly = c.split('_ONLY_')[1]
        configProhibits = true if configOnly && !testDescrip.index(configOnly)
      end
      next if configProhibits
      $FtpServers.each_index {|i|
        srvr = $FtpServers[i]
        @currentServer = srvr
        @total_tests += 1
        $stdout.write(("    Testing " + 
                test.sub('test_', '').sub(/ONLY_.*_ONLY_/, '') +
                "[#{i+1}]... ").ljust(42))
        $stdout.flush
        if send(test, srvr, @proxy)
          puts "Success"
        else
          puts "Failed - " + @error
          @errors << "#{description} " +
              "#{test.sub("test_", "")}[#{i+1}]  #{@error}"
          logout
        end
      }
    }
  end

end
